/* 
   FBRowHandler.m

   Copyright (C) 1999 MDlink online service center GmbH and Helge Hess

   Author: Helge Hess (helge.hess@mdlink.de)

   This file is part of the FB Adaptor Library

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; see the file COPYING.LIB.
   If not, write to the Free Software Foundation,
   59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
// $Id: FBRowHandler.m,v 1.2 1999/11/02 23:32:41 helge Exp $

#include "FBRowHandler.h"
#include "FBMetaData.h"
#include <FBCAccess/FBCAccess.h>
#import <Foundation/Foundation.h>
#import <eoaccess/EONull.h>

typedef struct {
  unsigned char blobType;
  unsigned int  blobSize;
  unsigned char *blobData;
} FBInlineBlob;

typedef struct {
  unsigned char blobType;
  unsigned char *blobHandle;
} FBHandleBlob;

@implementation FBRowHandler

static EONull *null = nil;

- (id)initWithData:(NSValue *)_rawData metaData:(FBMetaData *)_md {
  self->rows = fbcrhInitWith([_rawData pointerValue],
                             [_md cMetaData]);
  if (self->rows == NULL) {
    NSLog(@"Couldn't create row handler !");
    RELEASE(self);
    return nil;
  }
  self->metaData = RETAIN(_md);

  self->colCount   = fbcrhColumnCount(self->rows);
  self->rowCount   = fbcrhRowCount(self->rows);
  self->currentRow = 0;

  if (self->colCount > 0) {
    unsigned i;
    
    self->columnDataTypes = calloc(self->colCount, sizeof(unsigned));
    self->columnLabels    = calloc(self->colCount, sizeof(id));

    for (i = 0; i < self->colCount; i++) {
      FBCMetaData         *cm;
      FBCDatatypeMetaData *dmd;
      FBCColumnMetaData   *cmd;

      cm  = [self->metaData cMetaData];
      dmd = fbcmdDatatypeMetaDataAtIndex(cm, i);
      cmd = fbcmdColumnMetaDataAtIndex(cm, i);
      
      self->columnDataTypes[i] = fbcdmdDatatypeCode(dmd);
      self->columnLabels[i] =
        [NSString stringWithCString:fbccmdLabelName(cmd)];
    }
  }

  if (null == nil)
    null = RETAIN([EONull null]);
  
  return self;
}

- (void)dealloc {
  if (self->columnLabels) {
    unsigned i;
    for (i = 0; i < self->colCount; i++)
      RELEASE(self->columnLabels[i]);
    free(self->columnLabels);
    self->columnLabels = NULL;
  }
  if (self->columnDataTypes) {
    free(self->columnDataTypes);
    self->columnDataTypes = NULL;
  }
  if (self->rows) {
    fbcrhRelease(self->rows);
    self->rows = NULL;
  }
  RELEASE(self->metaData);
  [super dealloc];
}

- (void *)cRowHandler {
  return self->rows;
}

/* info */

- (unsigned)columnCount {
  return self->colCount;
}
- (int)rowCount {
  return self->rowCount;
}

- (FBMetaData *)metaData {
  return self->metaData;
}

/* rows */

static NSDictionary *_makeRow(FBRowHandler *self, void **row) {
  register unsigned i;
  NSDictionary *d;
  id values[self->colCount];
  
  if (row == NULL)
    return nil;

  for (i = 0; i < self->colCount; i++) {
    register void *value;

    value = row[i];

    if (value == NULL) {
      values[i] = null;
    }
    else {
      switch (self->columnDataTypes[i]) {
        case FB_Boolean:
          values[i] = [NSNumber numberWithBool:*(unsigned char *)value];
          break;
          
        case FB_SmallInteger:
        case FB_Integer:
          values[i] = [NSNumber numberWithInt:*(int *)value];
          break;
          
        case FB_Float:
          values[i] = [NSNumber numberWithFloat:*(float *)value];
          break;

        case FB_Double:
        case FB_Numeric:
        case FB_Decimal:
          values[i] = [NSNumber numberWithDouble:*(double *)value];
          break;

        case FB_Character:
        case FB_VCharacter:
          values[i] = [NSString stringWithCString:(unsigned char *)value];
          break;

        case FB_Date:
        case FB_Time:
        case FB_TimeTZ:
        case FB_Timestamp:
        case FB_TimestampTZ:
          values[i] = [NSString stringWithCString:(unsigned char *)value];
          /* convert to date .. */
          break;

        case FB_YearMonth:
          /*
            FB_YearMonth will return an integer (FB_Integer) and the value is
            the number of months in the interval.
            Use fbcdmdFrom() and fbcdmdTo() to format the value (if needed).
          */
          values[i] = [NSNumber numberWithInt:*(int *)value];
          break;

        case FB_DayTime:
          /*
            FB_DayTime will return a double (FB_Double) and the values is the
            number of seconds in the interval.
            Use fbcdmdFrom() and fbcdmdTo() to format the value (if needed).
          */
          values[i] = [NSNumber numberWithDouble:*(double *)value];
          break;

        case FB_CLOB:
          /*
            CLOB/BLOB are a bit more tricky, mainly because values of up to a
            given size are inlined in the fetch result, i.e. no need for a
            round trip to the server. 
          */
          if (*(unsigned char *)value == 0) {
            /* blob via handle */
            FBCBlobHandle *handle;
            unsigned      bsize;

            handle = fbcbhInitWithHandle(((FBHandleBlob *)value)->blobHandle);
            bsize  = fbcbhBlobSize(handle);

            values[i] = @"";
          }
          else {
            /* blob inline */
            FBInlineBlob *blob = value;
            
            values[i] = [NSString stringWithCString:blob->blobData
                                  length:blob->blobSize];
          }
          break;

        case FB_BLOB:
          /*
            CLOB/BLOB are a bit more tricky, mainly because values of up to a
            given size are inlined in the fetch result, i.e. no need for a
            round trip to the server. 
          */
          if (*(unsigned char *)value == 0) {
            /* blob via handle */
            FBCBlobHandle *handle;
            unsigned      bsize;

            handle = fbcbhInitWithHandle(((FBHandleBlob *)value)->blobHandle);
            bsize  = fbcbhBlobSize(handle);

            values[i] = [NSData data];
          }
          else {
            /* blob inline */
            FBInlineBlob *blob = value;

            values[i] = [NSData dataWithBytes:blob->blobData
                                length:blob->blobSize];
          }
          break;

        case FB_Bit:
        case FB_VBit:
          /*
            The next release of FrontBase has a small change in the way Bits
            are returned: 'value' will point to a structure with two fields: 

            struct BitValue 
              unsigned byteCount; 
              unsigned char* bytePtr; 
              ; 

            If you insert e.g. B'11110001' into a BIT column, the result is
            byteCount=1 and bytePtr points to a one byte array with the value
            0xF1.             
          */

        default:
          NSLog(@"unknown datatype %i ..", self->columnDataTypes[i]);
          values[i] = null;
          break;
      }
    }
  }

  return AUTORELEASE([[NSDictionary alloc] initWithObjects:values
                                           forKeys:self->columnLabels
                                           count:self->colCount]);
}

- (NSDictionary *)firstRow {
  void **row;

  if ((row = fbcrhFirstRow(self->rows)))
    return _makeRow(self, row);
  
  return nil;
}

- (NSDictionary *)nextRow {
  void **row;

  if ((row = fbcrhNextRow(self->rows)))
    return _makeRow(self, row);

  return nil;
}

- (NSDictionary *)rowAtIndex:(unsigned)_idx {
  void **row;

  if ((row = fbcrhRowAtIndex(self->rows, _idx)))
    return _makeRow(self, row);

  return nil;
}

/* raw rows */

- (void **)firstRawRow {
  return fbcrhFirstRow(self->rows);
}

- (void **)nextRawRow {
  return fbcrhNextRow(self->rows);
}

- (void **)rawRowAtIndex:(unsigned)_idx {
  return fbcrhRowAtIndex(self->rows, _idx);
}

/* enumerator */

- (id)nextObject {
  NSDictionary *obj;
  void **row;
  
  if (self->currentRow == 0) {
    row = fbcrhFirstRow(self->rows);
    obj = _makeRow(self, row);
  }
  else {
    row = fbcrhNextRow(self->rows);
    obj = _makeRow(self, row);
  }

  if (obj)
    (self->currentRow)++;

  return obj;
}

@end /* FBRowHandler */
